'
' Script to create Mesh Connectivity Layer (MCL) install kit.
'

'
' Read some environment variables.
'
set WshShell = CreateObject("WScript.Shell")
NTDrive = WshShell.ExpandEnvironmentStrings("%_NTDRIVE%")
DDKBuildEnv = WshShell.ExpandEnvironmentStrings("%DDKBUILDENV%")
if NTDrive = "%_NTDRIVE%" and DDKBuildEnv = "%DDKBUILDENV%" then
    WScript.Echo "Please run this command from a build environment shell"
    Wscript.Quit
end if
BuildAltDir = WshShell.ExpandEnvironmentStrings("%BUILD_ALT_DIR%")
if BuildAltDir = "%BUILD_ALT_DIR%" then
    BuildAltDir = ""
end if
BuildArch = WshShell.ExpandEnvironmentStrings("%_BuildArch%")
if BuildArch = "x86" then
    BuildArch = "i386"
end if
ObjDir = "obj" & BuildAltDir & "\" & BuildArch & "\"
MclRoot = WshShell.ExpandEnvironmentStrings("%MCLRoot%")
if MclRoot = "%MCLRoot%" then
    '
    ' If the user hasn't specified a root directory via an environment
    ' variable, we assume this script lives in the mcl subdirectory.
    '
    ScriptName = Wscript.ScriptFullName
    LlsrFolder = Left(ScriptName, InstrRev(ScriptName, "\"))
else
    if not Right(MclRoot, 1) = "\" then
        MclRoot = MclRoot & "\"
    end if
    MclFolder = MclRoot & "OpRouting\"
end if

'
' Source directories.
'
InfFolder = MclFolder & "inf\"
BinExeFolder = MclFolder & "exe\" & ObjDir
BinExpertFolder = MclFolder & "expert\" & ObjDir
BinParserFolder = MclFolder & "parser\" & ObjDir
BinSysFolder = MclFolder & "sys\" & ObjDir

'
' Destination directory.
' Check command-line arguments for a destination kit directory.
' If no kit directory is specified, we put it into a "MclKit"
' directory in the root of the drive used to build the kit.
'
set Args = Wscript.Arguments
if not Args.Count = 0 then
    KitFolder = Args(0)
    if not Right(KitFolder, 1) = "\" then
        KitFolder = KitFolder & "\"
    end if
else
    if NTDrive = "%_NTDRIVE%" then
        '
        ' If NTDrive isn't available, we'll use root of current drive.
        ' This will happen in the DDK build case.
        '
        NTDrive = ""
    end if
    KitFolder = NTDrive & "\MclKit\"
end if


'
' Dictionary listing of files to copy, and where to find them.
'
set Dict = CreateObject("Scripting.Dictionary")
Dict.Add "ReadMe.txt", MclFolder
Dict.Add "update.vbs", MclFolder
Dict.Add "license.txt", MclFolder
Dict.Add "mclmp.inf", InfFolder
Dict.Add "mcltp.inf", InfFolder
Dict.Add "mcl.exe", BinExeFolder
if DDKBuildEnv = "%DDKBUILDENV%" then
    '
    ' Expert only present in non-DDK builds.
    '
    Dict.Add "lqsrexp.dll", BinExpertFolder
end if
Dict.Add "lqsr.dll", BinParserFolder
Dict.Add "mcl.sys", BinSysFolder
Dict.Add "mcl.pdb", BinSysFolder


'
' Code to do the actual copying.
'
set FileSys = CreateObject("Scripting.FileSystemObject")
if FileSys.FolderExists(KitFolder) = False then
    FileSys.CreateFolder(KitFolder)
end if
for each Entry in Dict
    FileName = Dict.Item(Entry) & Entry
    if FileSys.FileExists(FileName) then
        '
        ' Copy the file, with overwrite flag to TRUE.
        '
        DestName = KitFolder & Entry
        if FileSys.FileExists(DestName) then
            set File = FileSys.GetFile(DestName)
            File.Attributes = File.Attributes and not 14000 and not 1
        end if
        set File = FileSys.GetFile(FileName)
        File.Copy DestName, TRUE
    else
        '
        ' Keep a list of files we wanted to copy but couldn't find.
        '
        Missing = Missing & FileName & vbLf
    end if
next

if not isEmpty(Missing) then
   ' WScript.Echo "Couldn't find following files to copy:" & vbLf & Missing
end if

if DDKBuildEnv = "%DDKBUILDENV%" then
    '
    ' Launch MakeHostsFile.vbs to create the hosts file.
    '
    WshShell.Run MclFolder & "..\Testbed\MakeHostsFile.vbs " & KitFolder, _
                 0, TRUE
end if
