// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#pragma warning(disable:4115) // named type definition in parentheses
#pragma warning(disable:4127) // conditional expression is constant
#pragma warning(disable:4200) // zero-sized array in struct/union
#pragma warning(disable:4201) // nameless struct/union
#pragma warning(disable:4213) // cast on l-value
#pragma warning(disable:4214) // bit field types other than int

#define NDIS_MINIPORT_DRIVER
#define NDIS51_MINIPORT 1
#define NDIS_WDM 1
#define NDIS51

//debug for KdPrint
#define debug 1

#undef NTDDI_VERSION
#define NTDDI_VERSION NTDDI_WINXP
#include <ndis.h>
#include <ntddk.h>
#include <malloc.h>

//
// Some warning disable commands must come after some of the system include
// files (e.g. ndis.h, ntddk.h) since the DDK versions re-enable them.
//
#pragma warning(disable:4200) // zero-sized array in struct/union

typedef struct ProtocolAdapter ProtocolAdapter;
typedef struct MiniportAdapter MiniportAdapter;
typedef struct NeighborCache NeighborCache;
typedef struct NeighborCacheEntry NeighborCacheEntry;
typedef struct RttParams RttParams; 
typedef struct PktPairParams PktPairParams; 
typedef struct EtxParams EtxParams; 
typedef struct WcettParams WcettParams; 
typedef struct UltParams UltParams;
typedef struct RecoveryCache RecoveryCache; 
typedef struct SRPacketHolder SRPacketHolder; 
typedef struct TimerCache TimerCache; 
typedef struct TimerHolder TimerHolder; 
typedef struct RecovTemp RecovTemp ; 
typedef struct MACLimitHolder MACLimitHolder ;
typedef struct MACLimit MACLimit ;
typedef struct ArbDijkstraCache ArbDijkstraCache ;


//
// Different rate limiting schemes. 
// Only one of these can be defined at a time!!
// These are deprecated and no longer supported.
#define SLIDING_RATE_LIMIT_ENABLED 0
#define AIMD_RATE_LIMIT_ENABLED	0


//if we want e2e rate-limit and the e2e rate-limit scheme that sends an e2e ACK
// based on the number of packets received (or a timeout), then want *both* of the
// following to be 1. This is the rate-control algorithm defined in the paper.
// We cannot have E2E_RATE_LIMIT_ENABLED = 0 and NEW_E2E_RATE_CONTROL = 1, but
// E2E_RATE_LIMIT_ENABLED = 1 and NEW_E2E_RATE_CONTROL = 0 is okay.
#define E2E_RATE_LIMIT_ENABLED 1
#define NEW_E2E_RATE_CONTROL 1

//debug for E2E ack
#define E2E_DEBUG 1

//should only be set to one if e2e rate-limiting is enabled. This will
// output the current acceptable rate to ttcp, which will then read the
// file and set its application layer datarate accordingly.
// This is deprecated, set to zero.
#define RATE_LIMIT_FILE 0
#define E2E_FILE_BUFFER_LENGTH 12
#define MAX_DATA_PKTS_PER_INTERVAL 117

//If we want to measure the link weights for some specified interval
// and then stop using them.
#define STOP_LINK_METRIC 1
//how long we measure link weights for (seconds)
// 1200 sec = 20 min
#define LINK_MEASURE_TIME ((Time) 1200)

#define CHANGELOGS      1       // Must be before internal headers.
#define NUM_LINKCHANGE_RECORDS  256
#define NUM_ROUTECHANGE_RECORDS 256

#define NDIS_STATUS_NO_RECOVERY_NODES 34532; 


#include "types.h"
#include "ether.h"
#include "lqsr.h"
#include "ntddmcl.h"
#include "sr.h"
#include "linkcache.h"
#include "reqtable.h"
#include "sendbuf.h"
#include "maintbuf.h"

//
// Put as 1 if we want to use the new method to update a packet state after 
// dequeuing from the preIFQ. 
// Recommended value: 1 (0 not tested in awhile)
#define NEW_PREIFQ_UPDATE_DEQUEUE 1

//
// Want to test new ack peeking against old ack peeking.
// Recommended value: 1 (0 not tested in awhile)
#define NEW_ACK_PEEK 1

//
// If we want piggy-acks to cancel the ack thinning timeout and reset 
// the ack skipping counter
// Recommended value: 1 (0 not tested in awhile)
#define PIGGY_ACKS_CANCEL 1

//
// If we haven't updated the fwding list in this long, then
// we will try to update the fwd list again.
//
#define UPDATE_FWD_LIST_TIME_DIFF ((Time) 2 * MILLISECOND)


//
// Put as 1 if we don't free a send packet to higher layer until all retransmissions are done
// 
// Put as 0 if we want to free a packet after the first time it was sent.
// Option 0 is highly recommended. (1 may not be stable)
#define OLD_NDIS_M_COMPLETE 0


//
// linkcache.c definitions
//

// This needs to be a positive integer, <= 100. It represents
// the percentage of our ETX to the next hop a forwarding node must beat.
// ETX(i, nexthop) < ETX (sender, nexthop)*CACHE_ETX_PERCENTAGE
//--now using for forwarding node selection
#define CACHE_ETX_PERCENTAGE 95

// How big of an ETX we allow for nodes to be in the forwarding list
// (from the sender). This is the default value.
// Remember that ETX is scaled by 4096
#define DEFAULT_ETX_FORWARD_SELECTION_THRESH (4 * 4096)

// Final equation
// we use to determine the ETX_FORWARD_THRESH is:
// max(ETX-to-default-next-hop*ETX_FACTOR, DEFAULT_ETX_FORWARD_SELECTION_THRESH)
//
#define ETX_FACTOR (4)

// How many forwarders we 
// limit ourselves too.
#define MAX_NUM_FORWARDERS (5)

// We try to pick forwarders until the loss link on the 
// virtual link reaches a certain threshold. This probability
// needs to be scaled by LOSS_SCALE (1000), so 1% loss = 0.01 = 10
//#define FORW_LOSS_THRESH 10
#define FORW_LOSS_THRESH 100

#define LOSS_SCALE 1000


//
// maintbuf.c definitions
//

//how long we hold an ACK before sending
#define ACK_THIN_TIMEOUT (30 * MILLISECOND)

//will skip 1 out of every NUM_ACKS_SKIPPED+1 ACKs
#define NUM_ACKS_SKIPPED 9

// E2E Ack thinning timeout
#define E2E_ACK_THIN_TIMEOUT (100 * MILLISECOND)

//send an E2E Ack for every 1 in E2E_NUM_ACKS_SKIPPED+1
// pkts received.
#define E2E_NUM_ACKS_SKIPPED 19

//maximum number of acks we'll put into a packet
#define NUM_ACK_THRESH 4

//the size of the local recovery caching entry
#define MAX_CACHE_ENTRY_SIZE 1000

//We want to send an E2E rate-control ACK for
// every NUM_E2E_SKIPS per-hop ACKs at the dest.
//(this is mostly used in protocol.c)
#define NUM_E2E_SKIPS 2

//For E2E rate-limiting, the minimum number of packets we can
// send in the interval (E2E_RATE_LIMIT_INTERVAL).
#define MIN_E2E_SENT 1

//how long the interval is for aimd rate limiting
#define E2E_RATE_LIMIT_INTERVAL ((Time) 200 * MILLISECOND)

//the additive increase for our rate control
#define E2E_ADDITIVE_CONSTANT 1


//
// recovery.c definitions
//


//
// We want exponential backoff for the restransmission timer.
// Previously, we had a backoff factor of 2, which meant we double the
// RTO at each retransmission of a packet. We want a fractional factor,
// so the new RTO factor is RESEND_EXP_FACTOR_NUM / RESEND_EXP_FACTOR_DEM
//
#define RESEND_EXP_FACTOR_NUM 3
#define RESEND_EXP_FACTOR_DEM 2


//
// number of milliseconds we set for the initial round trip
// timeout for retransmission timers.
//
#define INITIAL_RTO 40

//
//max number of retransmissions for local recovery 
//
#define MAX_NUM_ROUNDS 10

//How long it takes for the data to reach the next hop.
#define DATA_TX_TIME (45 * (MILLISECOND))

//the beta in the rttvar calculations
// scaled by 1000, so 1/4 = 0.25 should be 250
#define RTT_BETA (250)

//the alpha in the srtt calculations
// scaled by 1000, so 1/8 = 0.125 should be 125
#define RTT_ALPHA (125)

// the K in the srtt RTO calculations
#define RTO_K (4) 

// the G in the srtt RTO calculations
#define RTO_G (17 * (MILLISECOND))

//
// protocol.c definitions
//

//initial sequence number we assign packets before
// giving them a real sequence number
#define INITIAL_SEQ_NO ((uint) -1) 


//max size of our mac limiting queue
#define MAC_LIMIT_MAX_SIZE 50

//how many outstanding packets we allow in the interface queue
//#define MAX_MAC_OUTSTANDING 3
#define MAX_MAC_OUTSTANDING 3

//how many packets we look into the preIFQ for ACK peeking
#define ACK_PEEK_NUMBER 2

//
// This is for the limiting of the MAC queue.
//
struct MACLimitHolder {
	NDIS_PACKET		*Packet;	//pointer to the NDIS packet we'll try to send
	boolint			outstanding;// true if ProtocolTransmit has been sent on pkt 
								//and ProtocolTransmitComplete hasn't completed.
	
	MACLimitHolder	*next;		//next packet in our queue
	MACLimitHolder	*prev;		//prev packet in our queue
} ;

struct MACLimit {
	uint count ;					//number of packets we are holding
	MACLimitHolder *firstHolder;	//first packet holder
	MACLimitHolder *lastHolder;		//last packet holder
} ;

struct SRPacketHolder {
	SRPacketHolder	*next;		//next packet in our cache 
	SRPacketHolder	*prev;		//previous packet in our cache
	SRPacket		*srp ;		//the packet we are holding
	NDIS_PACKET		*Packet;	//pointer to the NDIS packet we'll make when we try to recover
	uint			count;		//the number of times this packet has been retransmitted
	Time			sendTime;	//the time the packet was sent (1st transmission), for srtt
	TimerHolder*	timerHolder ;	//the timer holder that is associated with this packet.
	uint			Type ;			//to indicate local recovery, priority-based forwarding, etc...
};

struct RecoveryCache {
	SRPacketHolder *firstHolder;	//doubly-linked list of all cached packets (the head)
	SRPacketHolder *lastHolder;		//the last packet holder in the LL (the tail)
	uint numPackets ;			//number of packets in the linked list
} ; 

struct TimerHolder {
	TimerHolder	*next; 
	TimerHolder	*prev; 
	Time			timeout ;					//expiration of this event
	SRPacketHolder* myPacketHolder ;			//the packet holder which is assoc. w/ this timerHolder.
	Time (*TimeoutCaller)(MiniportAdapter *VA, TimerHolder *tHolder);	//method we call when we have a timeout
};

struct TimerCache {
	TimerHolder *firstHolder;		//doubly-linked list of all timers (the head)
	TimerHolder *lastHolder;		//the last timer holder in the LL (the tail)
	uint numEvents ;				//number of timers in the linked list

} ; 

struct ArbDijkstraCache {
	//Time timeout ;				//time when caching entry is invalid //--let's just re-use LC->timeout
	VirtualAddress source ;		//source of flow
	VirtualAddress dest;		//dest of flow
	uint Metric;				//Metric we returned
	ArbDijkstraCache *next;		//next caching entry
	ArbDijkstraCache *prev;		//next caching entry
};

// If Count is not already aligned, then
// round Count up to an even multiple of "Pow2".  "Pow2" must be a power of 2.
//
// DWORD
// ROUND_UP_COUNT(
//     IN DWORD Count,
//     IN DWORD Pow2
//     );
#define ROUND_UP_COUNT(Count,Pow2) \
        ( ((Count)+(Pow2)-1) & (~(((LONG)(Pow2))-1)) )

// Minimum 5 milliseconds gap between broadcast packets.
#define MIN_BROADCAST_GAP (5 * MILLISECOND)

// Maximum number of broadcast packets we will queue.
#define MAX_FORWARD_QUEUE       16

// 100 ms backoff after first Route Request, 2 second max backoff
#define FIRST_BACKOFF (100 * MILLISECOND)        
#define MAX_BACKOFF (2 * SECOND)

// We think a link is broken if we do not receive an ack after this interval.
#define MAINTBUF_LINK_TIMEOUT   (500 * MILLISECOND)

// We retransmit an ack request after this interval.
// Must be <= MAINTBUF_LINK_TIMEOUT. Equal means no rexmits.
#define MAINTBUF_REXMIT_TIMEOUT (500 * MILLISECOND)

// We assume the link is working within this interval after an ack.
#define MAINTBUF_HOLDOFF_TIME   (250 * MILLISECOND)

// We delete a Maintenance Buffer Node after one day of inactivity.
// TODO - This is very large while we collect measurements.
#define MAINTBUF_IDLE_TIMEOUT   ((Time)24 * 60 * MINUTE)

// We limit the number of packets held in a Maintenance Buffer Node.
// This number should be large enough to hold "initial" packets
// sent when connections are starting up.
#define MAINTBUF_MAX_QUEUE      4

// We delay an Acknowledgement no longer than 80ms.
#define MAX_ACK_DELAY           (80 * MILLISECOND)

//ejr- We need to have immediate ACK's- not used anymore though....
//802.11a SIFS: 16 microseconds
//802.11b SIFS: 10 microseconds
//802.11g SIFS: 10 microseconds
//#define MAX_ACK_DELAY           (1 * TENTHMICRO)

#define SENDBUF_TIMEOUT (FIRST_BACKOFF + MAX_BACKOFF)

// We send link information at least every 10 seconds.
#define LINKINFO_PERIOD         ((Time)10 * SECOND)

// if PACKET_POOL_SZ <= SEND_BUF_SZ, then bad things happen really fast
#define PACKET_POOL_SZ       100
#define SEND_BUF_SZ          20

// Default Route Flap Damping Factor. Used in LinkCacheFlapDamp.
#define DEFAULT_ROUTE_FLAP_DAMPING_FACTOR 32

//rate-limit
//We want rate limiting to have 1 second intervals
#define SLIDING_RATE_LIMIT_TIMEOUT ((Time) 1 * SECOND)

//rate limit threshold
#define SLIDING_RATE_LIMIT_THRESH 1000

//value we scale percentages by for aimid rate control
#define AIMD_PROB_SCALE 1000

//how long the interval is for aimd rate limiting
#define AIMD_RATE_LIMIT_INTERVAL ((Time) 250 * MILLISECOND)

//the threshold of numSent/numAck, scalled by AIMD_PROB_SCALE
// So a value of 1.2 should be 1200
#define AIMD_RATE_LIMIT_THRESH 1200

//If we don't send enough packets, then don't decay
#define AIMD_PKT_SENT_THRESH 15

//by what factor we reduce our probably to send by
// for aimd rate limit
//want 1.2
#define AIMD_DECAY 1200
	
//by what constant we increase our probability to send
// for aimd rate limit. 1% is 10 (see AIMD_PROB_SCALE)
//want 5%
#define AIMD_INCREASE 50


//
// Functionality in sidebug.c.
//

#define MCL_POOL_TAG    'LCM'

#ifdef POOL_TAGGING

#ifdef ExAllocatePool
#undef ExAllocatePool
#endif

#define ExAllocatePool(type, size) ExAllocatePoolWithTag((type), (size), MCL_POOL_TAG)

#endif // POOL_TAGGING

#ifndef COUNTING_MALLOC
#define COUNTING_MALLOC DBG
#endif

#if COUNTING_MALLOC

#ifdef ExFreePool
#undef ExFreePool
#endif

#define ExAllocatePoolWithTag(type, size, tag)  CountingExAllocatePoolWithTag((type), (size), (tag), __FILE__, __LINE__)

#define ExFreePool(p) CountingExFreePool(p)

extern VOID *
CountingExAllocatePoolWithTag(
    IN POOL_TYPE        PoolType,
    IN ULONG            NumberOfBytes,
    IN ULONG            Tag,
    IN PCHAR            File,
    IN ULONG            Line);

extern VOID
CountingExFreePool(
    PVOID               p);

extern VOID
InitCountingMalloc(void);

extern VOID
DumpCountingMallocStats(void);

extern VOID
UnloadCountingMalloc(void);

#endif  // COUNTING_MALLOC


//
// Functionality in protocol.c.
//

//
// We can only use physical adapters that have a sufficiently large
// frame size that we can add our LQSR header.
//
#define PROTOCOL_MIN_FRAME_SIZE 1500

//
// We start dropping packets when a transmit queue gets bigger than this.
//
#define PROTOCOL_MAX_QUEUE      500

struct ProtocolAdapter {
    MiniportAdapter *VirtualAdapter;

    uint Index;
    GUID Guid;

    NDIS_HANDLE Handle;

    //
    // List of physical adapters bound to the virtual adapter,
    // locked by VirtualAdapter->Lock.
    //
    struct ProtocolAdapter **Prev, *Next;

    //
    // For allocating packets in ProtocolReceive.
    //
    NDIS_HANDLE PacketPool;
    NDIS_HANDLE BufferPool;

    //
    // For calling NdisOpenAdapter and NdisCloseAdapter.
    //
    KEVENT Event;
    NDIS_STATUS Status;

    //
    // Synchronizes Channel & Bandwidth attribute retrieval from NDIS.
    //
    FAST_MUTEX Mutex;

    //
    // Various attributes of the adapter.
    //
    boolint Use;                        // Should MCL use this adapter?
    uint MaxFrameSize;
    uint MaxTotalSize;
    PhysicalAddress Address;
    boolint Promiscuous;
    boolint ReceiveOnly;
    uint Channel;
    boolint ChannelConfigured;          // Was Channel manually configured?
    uint Bandwidth;
    boolint BandwidthConfigured;        // Was Bandwidth manually configured?

    //
    // Counters for monitoring/debugging purposes.
    //
    uint CountXmitLarge;        // Attempts to send a too-large packet.
    uint PacketsSent;           // Total calls to NdisSend.
    uint PacketsReceived;       // Total ProtocolReceive/ProtocolReceivePacket.
    uint PacketsReceivedTD;     // ProtocolReceive with Transfer Data.
    uint PacketsReceivedFlat;   // ProtocolReceive without Transfer Data.

    uint CountPacketPoolFailure;

    uint CountRecvOutstanding;  // Packets received but not completed.
    uint MaxRecvOutstanding;    // CountRecvOutstanding high-water mark.
    uint CountSentOutstanding;  // Packets sent but not completed.
    uint MaxSentOutstanding;    // CountSentOutstanding high-water mark.
};

extern ProtocolAdapter *
FindPhysicalAdapterFromIndex(MiniportAdapter *VA, uint Index);

extern ProtocolAdapter *
FindPhysicalAdapterFromGuid(MiniportAdapter *VA, const GUID *Guid);

extern boolint
ProtocolQueueFull(MiniportAdapter *VA, LQSRIf OutIf);

extern NDIS_STATUS
ProtocolInit(const WCHAR *Name, MiniportAdapter *VA);

extern void
ProtocolCleanup(MiniportAdapter *VA);

#if RATE_LIMIT_FILE
extern void
ProtocolCreateRateFile(uint dataPkts);
#endif

extern boolint ProtocolIsProbe(NDIS_PACKET *Packet);

extern void
ProtocolTransmit(ProtocolAdapter *PA, NDIS_PACKET *Packet);

extern void
ProtocolTransmitComplete(NDIS_HANDLE Handle, PNDIS_PACKET Packet, NDIS_STATUS Status)   ;

extern void
ProtocolRecoverTransmitComplete(NDIS_HANDLE Handle, PNDIS_PACKET Packet, NDIS_STATUS Status);

extern void 
updateTimeStruct(TimeStruct **ts, TimeStruct **last_ts, uint *count);

extern void 
pruneTimeStruct(TimeStruct **ts, TimeStruct **last_ts, uint *count); 

extern boolint 
shouldRateLimitDrop(MiniportAdapter *VA, AckCache *AC);

extern boolint 
shouldAIMDRateLimitDrop(MiniportAdapter *VA, AckCache *AC);

extern boolint
preIFQCanDelete(MiniportAdapter *VA, InternalAcknowledgement *Ack, SRPacket *srp, boolint higherETX) ;

extern void 
ProtocolRemoveFromPreIFQ(MiniportAdapter *VA, InternalAcknowledgement *Ack, boolint ackAlreadyUpdated, boolint higherETX);

extern void 
updateAIMDdeadline(AckCache *AC);

extern uint
updateCurAck(AckCache *AC, LQSRAckId seqNum, AckBitMap bitmap);

extern uint 
updateAIMDProb(AckCache *AC) ;

extern void 
updateE2Edeadline(AckCache *AC);

extern NDIS_STATUS 
ProtocolEnqueue(ProtocolAdapter *PA, NDIS_PACKET *Packet);

extern void 
ProtocolDequeue(ProtocolAdapter *PA, NDIS_PACKET *Packet);

extern void 
ProtocolCheckQueue(ProtocolAdapter *PA);

extern boolint 
protocolCanPeek(MiniportAdapter *VA, NDIS_PACKET *Packet);

extern boolint 
protocolIsACK(MiniportAdapter *VA, NDIS_PACKET *Packet);

extern void 
protocolUpdateAckState(MiniportAdapter *VA, NDIS_PACKET *Packet, KIRQL OldIrql);

extern NDIS_PACKET*
protocolUpdateAckAndFwd(MiniportAdapter *VA, NDIS_PACKET *Packet, KIRQL OldIrql);

extern void
ProtocolRecoverTransmit(ProtocolAdapter *PA, NDIS_PACKET *Packet);

#if E2E_RATE_LIMIT_ENABLED
extern void 
ProtocolSendE2EAck (AckCache *AC, uint numAcked) ;
#endif

extern void
ProtocolForwardRequest(MiniportAdapter *VA, SRPacket *srp);

extern void
ProtocolResetStatistics(ProtocolAdapter *PA);

extern void
ProtocolQuery(ProtocolAdapter *PA, MCL_INFO_PHYSICAL_ADAPTER *Info);

extern NTSTATUS
ProtocolControl(ProtocolAdapter *PA, MCL_CONTROL_PHYSICAL_ADAPTER *Control);

extern NTSTATUS
ProtocolPersistControl(ProtocolAdapter *PA,
                       MCL_CONTROL_PHYSICAL_ADAPTER *Control);

extern NDIS_STATUS
ProtocolRequestHelper(ProtocolAdapter *PA, NDIS_REQUEST_TYPE RT, NDIS_OID OID,
                      void *Info, uint Length, uint *Needed);


//
// Functionality in neighbor.c.
//

//
// The NeighborCacheEntry maps a virtual address and
// an interface index (representing another node's interface)
// to a physical address (of that other interface).
//
struct NeighborCacheEntry {
    NeighborCacheEntry *Next;
    NeighborCacheEntry *Prev;

    VirtualAddress VAddress;
    LQSRIf InIf;
    PhysicalAddress PAddress;
};

struct NeighborCache {
    KSPIN_LOCK Lock;
    NeighborCacheEntry *FirstNCE;
    NeighborCacheEntry *LastNCE;
};

__inline NeighborCacheEntry *
SentinelNCE(NeighborCache *NC)
{
    return (NeighborCacheEntry *) &NC->FirstNCE;
}

extern void
NeighborCacheInit(NeighborCache *NC);

extern void
NeighborCacheCleanup(NeighborCache *NC);

extern void
NeighborCacheFlushAddress(NeighborCache *NC, const VirtualAddress VAddr,
                          LQSRIf InIf);

extern void
NeighborReceivePassive(NeighborCache *NC, const VirtualAddress VAddr,
                       LQSRIf InIf, const PhysicalAddress PAddr);

extern boolint
NeighborFindPhysical(NeighborCache *NC, const VirtualAddress VAddr,
                     LQSRIf InIf, PhysicalAddress PAddr);


//
// Functionality in pback.c
//
// Currently the piggy-back cache uses a simple sorted list
// of options waiting to be sent. We expect the cache size
// to be small - in a lightly-loaded system not many options
// are sent, and in a highly-loaded system many packets are
// sent and so options leave the cache quickly.
//

#define PBACK_ROUTE_REPLY_TIMEOUT       5000            // 1/2 millisecond.
#define PBACK_ROUTE_ERROR_TIMEOUT       50000           // 5 milliseconds.
#define PBACK_INFO_REPLY_TIMEOUT        5000000         // 1/2 second.
#define PBACK_ROUTE_REPLY_SR_TIMEOUT    10000000        // 1 second.

typedef struct PbackOption PbackOption;

struct PbackOption {
    PbackOption *Next;
    VirtualAddress Dest;
    InternalOption *Opt;
    Time Timeout;
};

typedef struct PbackCache {
    KSPIN_LOCK Lock;
    PbackOption *List;

    uint Number;
    uint HighWater;
    Time AckMaxDupTime;

    uint CountPbackTooBig;
    uint CountPbackTotal;
    uint CountAloneTotal;
    uint CountPbackAck;
    uint CountAloneAck;
    uint CountPbackReply;
    uint CountAloneReply;
    uint CountPbackError;
    uint CountAloneError;
    uint CountPbackInfo;
    uint CountAloneInfo;
} PbackCache;

extern void
PbackInit(MiniportAdapter *VA);

extern void
PbackCleanup(MiniportAdapter *VA);

extern void
PbackResetStatistics(MiniportAdapter *VA);

extern void
PbackSendOption(MiniportAdapter *VA, const VirtualAddress Dest,
                InternalOption *Opt, Time Timeout);

extern void
PbackSendPacket(MiniportAdapter *VA, SRPacket *SRP);

extern Time
PbackTimeout(MiniportAdapter *VA, Time Now);

extern void
PbackSendComplete(MiniportAdapter *VA, SRPacket *SRP, NDIS_STATUS Status);

extern uint
PbackPacketSize(SRPacket *SRP);


//
// Functionality in etx.c.
//

struct EtxParams {
    uint ProbePeriod; 
    uint LossInterval;
    uint Alpha;
    uint PenaltyFactor;
    Time ProbeTimeout;
}; 

extern uint 
bestETX (VirtualAddress dest, MiniportAdapter *VA) ;

extern uint 
getETX (VirtualAddress src, VirtualAddress dest, MiniportAdapter *VA); 

extern void EtxInit(
    MiniportAdapter *VA);


//
// Functionality in wcett.c
//

struct WcettParams {
    uint ProbePeriod; 
    uint LossInterval;
    uint Alpha;
    uint PenaltyFactor;
    Time ProbeTimeout;
    uint Beta;
    uint PktPairProbePeriod;
    uint PktPairMinOverProbes;
};

extern void
WcettInit(MiniportAdapter *VA);

extern void
WcettReceivePktPairProbeReply(MiniportAdapter *VA,
                              InternalProbeReply *ProbeReply);

extern uint
WcettEncodeBandwidth(uint Bandwidth);

extern uint
WcettDecodeBandwidth(uint Bandwidth);


//
// Functionality in pktpair.c.
//

struct PktPairParams {
    uint Alpha; 
    uint ProbePeriod; 
    uint PenaltyFactor;
}; 

extern void 
PktPairInit(MiniportAdapter *VA);

void
PktPairSendProbeReply(
    MiniportAdapter *VA, 
    InternalProbe *Probe, 
    Time Now, 
    Time OutDelta);

void
PktPairReceiveProbeReply(
    MiniportAdapter *VA,
    InternalProbeReply *ProbeReply);

NDIS_STATUS
PktPairCreateProbePacket(
    MiniportAdapter *VA,
    Link *Adjacent,
    Time Timestamp,
    NDIS_PACKET **ReturnPacket,
    uint Seq,
    boolint LargeProbe);

void
PktPairSendProbeComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status);


//
// Functionality in rtt.c
//

struct RttParams {
    uint Alpha; 
    uint ProbePeriod; 
    uint HysteresisPeriod; 
    uint PenaltyFactor;
    uint SweepPeriod;
    Time SweepTimeout;
    boolint Random;
    boolint OutIfOverride;
}; 

extern void
RttInit(MiniportAdapter *VA); 

extern void
RttReceiveProbeReply(MiniportAdapter *VA, InternalProbeReply *ProbeReply);

extern void 
RttSelectOutIf(
    MiniportAdapter *VA,
    const uchar *Dest,
    LQSRIf *OutIf,
    LQSRIf *InIf);


//
// Functionality in ult.c.
//

struct UltParams {
    Time UcastProbeTimeout;
    uint UcastProbePeriod;
    uint UcastLowThreshold;     // Scaled by MAXPERCENT.
    uint UcastHighThreshold;    // Scaled by MAXPERCENT.
    uint UcastLossPenalty;      // Scaled by MAXPERCENT.
    Time BcastProbeTimeout;
    uint BcastProbePeriod;
    uint BcastLossThreshold;
};

extern void UltInit(
    MiniportAdapter *VA);

extern void
UltSendUcastProbeComplete(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status);

//
// Functionality in crypto.c.
//

#define AES_KEYSIZE_128 (16)

#define CRYPTO_KEY_MAC_LENGTH   16

extern void
CryptoKeyMACModify(MiniportAdapter *VA, uchar Key[LQSR_KEY_SIZE]);

extern void
CryptoMAC(uchar CryptoKeyMAC[CRYPTO_KEY_MAC_LENGTH],
          uchar *MAC, uint Length,
          NDIS_PACKET *Packet, uint Offset);

extern NDIS_STATUS
CryptoEncryptPacket(MiniportAdapter *VA, NDIS_PACKET *OrigPacket,
                    uint Offset,
                    uchar IV[LQSR_IV_LENGTH], NDIS_PACKET **pEncryptedPacket);

extern NDIS_STATUS
CryptoDecryptPacket(MiniportAdapter *VA, NDIS_PACKET *OrigPacket,
                    uint OrigOffset, uint DecryptOffset,
                    uchar IV[LQSR_IV_LENGTH], NDIS_PACKET **pDecryptedPacket);

//
// NDIS-related helper functions, in ndishack.c.
//

typedef struct PacketContext {
    union {
        void (*TransmitComplete)(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status);
        void (*ReceiveComplete)(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status);
    };
    NDIS_PACKET *OrigPacket;
    NDIS_BUFFER *OrigBuffer;
    void *CloneData;
    union {
        SRPacket *srp;
        MaintBufPacket *MBP;
    };
    ProtocolAdapter *PA;
	uint queueID;			//identifier for protocol queue and dequeue
} PacketContext;

__inline PacketContext *
PC(NDIS_PACKET *Packet)
{
    return (PacketContext *)Packet->ProtocolReserved;
}

extern NDIS_STATUS
NdisClonePacket(NDIS_PACKET *OrigPacket,
                NDIS_HANDLE PacketPool, NDIS_HANDLE BufferPool,
                uint OrigHeaderLength, uint CloneHeaderLength,
                uint LookAhead,
                void **pOrigHeader,
                NDIS_PACKET **pClonePacket, void **pCloneHeader);

//ejr
extern 
NDIS_STATUS
CopyWholePacket(MiniportAdapter *VA,
	NDIS_PACKET *OrigPacket,
    NDIS_PACKET **pClonePacket,
    NDIS_HANDLE PacketPool,
    NDIS_HANDLE BufferPool,
	void **dataPtr);


extern void
NdisFreePacketClone(NDIS_PACKET *ClonePacket);

extern NDIS_STATUS
NdisMakeEmptyPacket(NDIS_HANDLE PacketPool, NDIS_HANDLE BufferPool,
                    uint Size,
                    NDIS_PACKET **pPacket, void **pData);

//
// Hack around problems and limitations in the NDIS interface,
// by relying on knowledge of internal NDIS data structures.
//

__inline PNDIS_BUFFER
NdisFirstBuffer(PNDIS_PACKET Packet)
{
    return Packet->Private.Head;
}

extern void
NdisAdjustBuffer(NDIS_BUFFER *Buffer, void *Address, uint Length);

extern NDIS_HANDLE
NdisProtocolFromBindContext(NDIS_HANDLE BindContext);

extern void **
NdisReservedFieldInProtocol(NDIS_HANDLE Protocol);


//
// Functionality in miniport.c.
//

typedef union MetricParams {
    RttParams Rtt;
    PktPairParams PktPair;
    EtxParams Etx;
    WcettParams Wcett;
    UltParams Ult;
} MetricParams; 

extern struct MiniportAdapters {
    KSPIN_LOCK Lock;
    MiniportAdapter *VirtualAdapters;
} MiniportAdapters;

struct MiniportAdapter {
    MiniportAdapter **Prev, *Next;

    uint Index;
    GUID Guid;

    NDIS_HANDLE MiniportHandle;
    NDIS_HANDLE ProtocolHandle;

    PDEVICE_OBJECT PhysicalDeviceObject;
    PDEVICE_OBJECT DeviceObject;

    KSPIN_LOCK Lock;
    ProtocolAdapter *PhysicalAdapters;
    uint NextPhysicalAdapterIndex;

    Time Timeout;
    KTIMER Timer;
    KDPC TimeoutDpc;
    uint TimeoutMinLoops;
    uint TimeoutMaxLoops;

    NeighborCache NC;
    PbackCache PCache;
    LinkCache *LC;
    ForwardedRequestTable *ReqTable;
    SendBuf *SB;
    MaintBuf *MaintBuf;


    SRPacket *ForwardList;
    SRPacket **ForwardLast;
    Time ForwardTime;

    Time LastLinkInfo;

    VirtualAddress Address;

    boolint Snooping;           // Learn route info from overhead packets?
    
    boolint Crypto;             // Encrypt the packets? 
    uchar CryptoKeyMAC[CRYPTO_KEY_MAC_LENGTH];
    uchar CryptoKeyAES[AES_KEYSIZE_128];

    boolint ArtificialDrop;     // Simulate lossy links by dropping packets?

    Time LinkTimeout;           // How long do stale links live?

    NDIS_MEDIUM Medium;
    uint PacketFilter;
    uint LookAhead;
    uint MediumLinkSpeed;
    uint MediumMinPacketLen;
    uint MediumMaxPacketLen;
    uint MediumMacHeaderLen;
    uint MediumMaxFrameLen;

    NDIS_HANDLE PacketPool;
    NDIS_HANDLE BufferPool;
    uint CountPacketPoolFailure;

    //
    // Parameters for Metric being used. 
    // 

    MetricType MetricType;
    MetricParams MetricParams;
    uint RouteFlapDampingFactor;   // Route flap damping factor. 0 means no damping.

    boolint (*IsInfinite)(uint LinkMetric);
    boolint (*IsBad)(uint LinkMetric);
    uint (*ConvMetric)(uint LinkMetric);
    void (*InitLinkMetric)(MiniportAdapter *VA,
                           int SNode, Link *Link, Time Now);
    uint (*PathMetric)(MiniportAdapter *VA, Link **Hops, uint NumHops);
    void (*Penalize)(MiniportAdapter *VA, Link *Adj);
    void (*ReceiveProbe)(MiniportAdapter *VA, InternalProbe *Probe,
                         LQSRIf InIf);
    Time (*MetricTimeout)(MiniportAdapter *VA, Time Now);

    //
    // Counters for monitoring/debugging purposes.
    //
    uint CountXmit;
    uint CountXmitSmall;
    uint CountXmitLarge;
    uint CountXmitLocally;
    uint CountXmitQueueFull;
    uint CountRexmitQueueFull;
    uint CountXmitNoRoute;
    uint CountXmitMulticast;
    uint CountXmitRouteRequest;
    uint CountXmitRouteReply;
    uint CountXmitRouteError;
    uint CountXmitRouteErrorNoLink;
    uint CountXmitSendBuf;
    uint CountXmitMaintBuf;
    uint CountXmitForwardUnicast;
    uint CountXmitForwardQueueFull;
    uint CountXmitForwardBroadcast;
    uint CountXmitInfoRequest;
    uint CountXmitInfoReply;
    uint CountXmitProbe;
    uint CountXmitProbeReply;
    uint CountXmitLinkInfo;

    uint CountLinkInfoTooManyLinks;
    uint CountLinkInfoTooLarge;

    uint CountSalvageAttempt;
    uint CountSalvageStatic;
    uint CountSalvageOverflow;
    uint CountSalvageNoRoute;
    uint CountSalvageSameNextHop;
    uint CountSalvageTransmit;
    uint CountSalvageQueueFull;

    uint CountRecv;
    uint CountRecvLocally;
    uint CountRecvLocallySalvaged;
    uint CountRecvBadMAC;
    uint CountRecvRouteRequest;
    uint CountRecvRouteReply;
    uint CountRecvRouteError;
    uint CountRecvAckRequest;
    uint CountRecvAck;
    uint CountRecvSourceRoute;
	uint CountRecvRecovery ;
    uint CountRecvInfoRequest;
    uint CountRecvInfoReply;
    uint CountRecvDupAckReq;
    uint CountRecvProbe;
    uint CountRecvProbeReply;
    uint CountRecvLinkInfo;
    uint CountRecvRecursive;
    uint CountRecvEmpty;
    uint CountRecvSmall;
    uint CountRecvLarge;
    uint CountRecvDecryptFailure;

    uint MinVersionSeen;
    uint MaxVersionSeen;

    uint TotalForwardingStall;          // Microseconds.

    uint ForwardNum;
    uint ForwardMax;
    uint CountForwardFast;
    uint CountForwardQueue;
    uint CountForwardDrop;

	//ejr
	// Additional info for Local Recovery
	//	
    KSPIN_LOCK RecoveryLock;
	RecoveryCache myCache; 
	TimerCache timerCache; 
	
	Time RecoveryTimeout;
    KTIMER RecoveryTimer;
    KDPC RecoveryTimeoutDpc;

	//ejr
	AckCacheList acl ;			//this lists the (src,dest) pairs that we have an AckCache for. 
								//it will be stored at src, so it is a list of all destinations.

	MACLimit mlq	;			//this is the mac-limiting queue

	uint queueCounter;			//keep track of ID's for this VA in protocol queue and dequeue

	//just some junk so SendACKComplete != PbackSendComplete
	uint junk;

	//added some stuff to give etx probes a seqnum
	uint etx_seqno ; 

	//want to have some caching for arb_dijkstra
	ArbDijkstraCache *adc ; //ArbDijkstraCache
	uint num_adcs ;			//number of above caching entries 

#if E2E_RATE_LIMIT_ENABLED	
	uint numE2EAcksSent; 
	uint numE2EAcksRecv; 
#endif

#if STOP_LINK_METRIC
	Time start_link_time;
#endif
};

extern MiniportAdapter *
FindVirtualAdapterFromIndex(uint Index);

extern MiniportAdapter *
FindVirtualAdapterFromGuid(const GUID *Guid);

extern boolint
destinationShouldReceive(MiniportAdapter *VA, SourceRoute *SR) ;

extern NDIS_STATUS 
updateForwardInfo(MiniportAdapter *VA, SRPacket *srp) ;

extern boolint
OnForwardingList(MiniportAdapter *VA, VirtualAddress *Addr, SourceRoute *SR) ;

extern boolint
MiniportInit(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath);

extern void
MiniportReceivePacket(MiniportAdapter *VA, ProtocolAdapter *PA,
                      NDIS_PACKET *Original,
                      void (*FreePacket)(ProtocolAdapter *PA,
                                         NDIS_PACKET *Packet));

extern void
MiniportIndicateStatusConnected(MiniportAdapter *A);

extern void
MiniportIndicateStatusDisconnected(MiniportAdapter *A);

extern void
MiniportReceiveComplete(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status);

extern void
MiniportSendComplete(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status);

extern void
MiniportRetransmitSendComplete(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status);

extern void 
MiniportForwardComplete(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status);

extern void 
MiniportSendRouteRequest(MiniportAdapter *VA,
                         const VirtualAddress Target, LQSRReqId Identifier);

extern void
MiniportSendViaRouteRequest(MiniportAdapter *VA,
                            SRPacket *SRP,
                            void (*Complete)(MiniportAdapter *VA,
                                             SRPacket *SRP,
                                             NDIS_STATUS Status));

extern void 
MiniportSendRouteError(MiniportAdapter *VA, const SRPacket *SRP);

extern void
MiniportSendInfoRequest(MiniportAdapter *VA);

extern void
MiniportResetStatistics(MiniportAdapter *VA);

extern void
MiniportRescheduleTimeout(MiniportAdapter *VA, Time Now, Time Timeout);

extern void
MiniportRescheduleRecoveryTimeout(MiniportAdapter *VA, Time Now, Time Timeout);


extern void
MiniportProbeComplete(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status);

NDIS_STATUS
MiniportWriteMetricType(NDIS_HANDLE Config, MetricType MetricType);

NDIS_STATUS
MiniportReadMetricType(NDIS_HANDLE Config, MetricType *MetricType);

__inline void
MiniportAllocatePacket(
    MiniportAdapter *VA,
    NDIS_PACKET **pPacket,
    NDIS_STATUS *pStatus)
{
    NdisAllocatePacket(pStatus, pPacket, VA->PacketPool);
    if (*pStatus != NDIS_STATUS_SUCCESS)
        InterlockedIncrement((PLONG)&VA->CountPacketPoolFailure);
}

__inline NDIS_STATUS
MiniportClonePacket(
    MiniportAdapter *VA,
    NDIS_PACKET *OrigPacket,
    uint OrigHeaderLength, uint CloneHeaderLength,
    uint LookAhead,
    void **pOrigHeader,
    NDIS_PACKET **pClonePacket, void **pCloneHeader)
{
    NDIS_STATUS Status;

    Status = NdisClonePacket(OrigPacket, VA->PacketPool, VA->BufferPool,
                             OrigHeaderLength, CloneHeaderLength,
                             LookAhead, pOrigHeader,
                             pClonePacket, pCloneHeader);
    if (Status != NDIS_STATUS_SUCCESS)
        InterlockedIncrement((PLONG)&VA->CountPacketPoolFailure);
    return Status;
}



__inline NDIS_STATUS
MiniportMakeEmptyPacket(
    MiniportAdapter *VA, uint Size,
    NDIS_PACKET **pPacket, void **pData)
{
    NDIS_STATUS Status;

    Status = NdisMakeEmptyPacket(VA->PacketPool, VA->BufferPool,
                                 Size, pPacket, pData);
    if (Status != NDIS_STATUS_SUCCESS)
        InterlockedIncrement((PLONG)&VA->CountPacketPoolFailure);
    return Status;
}

extern boolint
MiniportIsInfinite(uint LinkMetric);

extern boolint
MiniportIsBad(uint LinkMetric);

extern uint
MiniportConvMetric(uint LinkMetric);

extern uint
MiniportPathMetric(MiniportAdapter *VA, Link **Hops, uint NumHops);

extern void 
MiniportInitLinkMetric(MiniportAdapter *VA, int SNode, Link *Link, Time Now);

extern void
MiniportForwardProbe(MiniportAdapter *VA, SRPacket *srp);

extern NTSTATUS
MiniportPersistControl(MiniportAdapter *VA,
                       MCL_CONTROL_VIRTUAL_ADAPTER *Control);

extern void
MiniportReadControl(MiniportAdapter *VA,
                    MCL_CONTROL_VIRTUAL_ADAPTER *Control);

extern NTSTATUS
MiniportControl(MiniportAdapter *VA, MCL_CONTROL_VIRTUAL_ADAPTER *Control);


//
// Functionality in recoverycache.c
//

extern uint 
forwardListSize(SourceRoute *SR);

extern void
addTimer (TimerHolder *newth, MiniportAdapter *VA);

extern void
addPacket(SRPacketHolder *newph, MiniportAdapter *VA);

extern void
deletePacket (SRPacketHolder *curHolder, MiniportAdapter *VA);

extern void
deleteTimer (TimerHolder *curHolder, MiniportAdapter *VA);

extern void
updateTimer(TimerHolder *curHolder, Time NewTime, MiniportAdapter *VA);

extern 
boolint HigherForwardPriority(MiniportAdapter *VA, SourceRoute *sr, VirtualAddress *first) ;

extern
boolint HigherETXPriority(MiniportAdapter *VA, VirtualAddress *Fwder, VirtualAddress *Dest);

extern void
RecoveryTimer(PVOID VVA) ;

extern void 
LocalRecoveryHalt(MiniportAdapter *VA);

extern Time
LocalRecoveryCheckRecovery(MiniportAdapter *VA) ;

extern void
LocalRecoveryInitialize(MiniportAdapter *VA) ; 

extern boolint 
LocalRecoveryProcessAck(MiniportAdapter *VA, InternalAcknowledgement *Ack, boolint *higherETX) ;

extern Time 
LocalRecoveryCheckRetransmission(MiniportAdapter *VA, TimerHolder *tHolder);

extern boolint 
LocalRecoveryAddRetransmission(MiniportAdapter *VA, SRPacket *srp, MaintBufNode *MBN) ;

extern void 
LocalRecoveryAddForwarding(MiniportAdapter *VA, SRPacket *srp, int my_slot) ; 

extern Time 
LocalRecoveryCheckForwarding(MiniportAdapter *VA, TimerHolder *tHolder);

extern void 
LocalRecoveryUpdateRetransmissionTimer(MiniportAdapter *VA, SRPacket *srp) ;

//
// Functionality in ackcache.c
//
extern void 
AckCacheFillAck(AckCache *ac, LQSRAckId *fillSeqNum, AckBitMap *fillBitmap, LQSRAckId rcvdSeqNum) ;

extern boolint 
AckCacheDeletePacket(InternalAcknowledgement *Ack, LQSRAckId storeSeqNum);

extern void 
AckCacheShiftRight(AckBitMap *abm, int shift);

extern void 
AckCacheShiftLeft(AckBitMap *abm, int shift);

extern void 
AckCacheSetBit(AckBitMap *abm, int index);

extern boolint 
AckCacheCheckBit(AckBitMap *abm, int index) ;

extern LQSRAckId 
AckCacheAdjustBitMap(AckBitMap *abm, LQSRAckId oldAck, LQSRAckId newAck);

extern void 
AckCacheUpdateAckState(AckCache *AC, InternalAcknowledgement *Ack);


//
// Functionality in rtt.c
//

extern void
RttReceiveProbeReply(MiniportAdapter *VA, InternalProbeReply *ProbeReply);

extern void
RttSendProbeComplete(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status);


//
// Functionality in driver.c and io.c.
//

extern uint Version;

extern DEVICE_OBJECT *OurDeviceObject;

extern PDRIVER_DISPATCH IoMajorFunctions[IRP_MJ_MAXIMUM_FUNCTION + 1];

extern NTSTATUS
IoCreate(PDEVICE_OBJECT DeviceObject, PIRP Irp);

extern NTSTATUS
IoCleanup(PDEVICE_OBJECT DeviceObject, PIRP Irp);
extern NTSTATUS
IoClose(PDEVICE_OBJECT DeviceObject, PIRP Irp);

extern NTSTATUS
IoControl(PDEVICE_OBJECT DeviceObject, PIRP Irp);


//
// Functionality in random.c.
//

extern int
RandomInit(void);

extern void
GetRandom(uchar *Buffer, uint Length);

extern uint
GetRandomNumber(uint Max);

extern NDIS_STATUS
GetSystemRandomBits(uchar *Buffer, uint Length);


//
// Functionality in reg.c.
//

typedef enum {
    OpenRegKeyRead,
    OpenRegKeyCreate,
    OpenRegKeyDeleting
} OpenRegKeyAction;

extern NTSTATUS
OpenRegKey(PHANDLE HandlePtr, HANDLE Parent, const WCHAR *KeyName,
           OpenRegKeyAction Action);

extern NTSTATUS
GetRegDWORDValue(HANDLE KeyHandle, const WCHAR *ValueName, PULONG ValueData);

extern NTSTATUS
SetRegDWORDValue(HANDLE KeyHandle, const WCHAR *ValueName,
                 IN ULONG ValueData);

extern NTSTATUS
GetRegNetworkAddress(HANDLE KeyHandle, const WCHAR *ValueName,
                     OUT VirtualAddress Address);

extern NTSTATUS
SetRegNetworkAddress(HANDLE KeyHandle, const WCHAR *ValueName,
                     IN VirtualAddress Address);

extern NTSTATUS
GetRegGuid(HANDLE KeyHandle, const WCHAR *ValueName,
           OUT GUID *Guid);

extern NTSTATUS
GetRegBinaryValue(HANDLE KeyHandle, const WCHAR *ValueName,
                  OUT uchar *Buffer, uint Length);

extern NTSTATUS
SetRegBinaryValue(HANDLE KeyHandle, const WCHAR *ValueName,
                  IN uchar *Buffer, uint Length);

//
// Define missing interlocked functions.
//

__inline LONGLONG
InterlockedRead64(LONGLONG volatile *Target)
{
#if defined(_M_IA64) || defined(_M_AMD64)
    return *Target;
#else
    LONGLONG Value = 0;
    return InterlockedCompareExchange64(Target, Value, Value);
#endif
}

__inline void
InterlockedWrite64(LONGLONG volatile *Target, LONGLONG Value)
{
#if defined(_M_IA64) || defined(_M_AMD64)
    *Target = Value;
#else
    LONGLONG OldValue = InterlockedRead64(Target);
    (void) InterlockedCompareExchange64(Target, Value, OldValue);
#endif
}

__inline void
InterlockedIncrementHighWater(LONG volatile *Count, LONG volatile *HighWater)
{
    LONG NewValue = InterlockedIncrement(Count);
    LONG OldHighWater;
    while ((OldHighWater = *HighWater) < NewValue) {
        if (InterlockedCompareExchange(HighWater, NewValue, OldHighWater) == NewValue)
            break;
    }
}
