// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

typedef struct MaintBuf MaintBuf;
typedef struct MaintBufPacket MaintBufPacket;
typedef struct MaintBufNode MaintBufNode;
typedef struct AckCache AckCache ; 
typedef struct AckCacheList AckCacheList ; 
typedef struct CachingEntry CachingEntry; 
typedef struct TimeStruct TimeStruct; 

struct TimeStruct {
	TimeStruct *next;
	TimeStruct *prev; 
	Time timestamp; 
};

//static array of size MAX_CACHE_ENTRY_SIZE.
struct CachingEntry {
	LQSRAckId	SeqNum ;		//the seqNum of the packet we cached
	VirtualAddress forwarder;	//the node that forwarded the packet
};


struct AckCache {	
	AckCache *next;			//linked list of AckCaches

	VirtualAddress source;	//the source of the flow...should be the same address as MBN it is stored in.
	VirtualAddress dest ;	//the destination of the flow...source is the MBN the AckCache is stored at.
	
	VirtualAddress to ;		//who we should send the ACK to.
	LQSRIf InIf ;
	LQSRIf OutIf;

	//bitmap of packets we've received for this tuple. Rightmost bitmap entry
	//represents AckNum+1, leftmost represents AckNum+bit_map+size
	AckBitMap	bitmap;		
	LQSRAckId	AckNum;		//cumulative ack (highest #'d ACK we've consecutively received)
	ushort NumAckSkipped;		//the number of times we've consecutively skipped an ACK (for ACK thinning)
	KTIMER AckTimer;			//allows us to set the timeout for ACK thinning 
	KDPC AckTimeoutDpc;			//for firing the timer
	MiniportAdapter *VA;		//needed for ACK timeout method, initialized in maintbuffindnode()

	uint nextSeqNumToSend ;		//for sender to determine next seqNum to use.

	//we want to store the RTO for a given flow, so let's stick it in here
	Time srtt ;        //need to still initialize these
	Time rttvar; 
	boolint firstRTT;	//need to know if it is first RTT measurement or not

	//for cross-flow ACKs, we need to know the number of outstanding 
	// packets that need to be ACKed.
	uint numOutAckRequest;	//number of packets that need to be acked for this flow

#if E2E_RATE_LIMIT_ENABLED		
	//dest
	uint NumAckSent;		//the total number of ACKs sent on the per-hop link (used for E2E rate control)
							//this is for the destinaition of a flow

	//src
	uint curNumSent;	//number of packet sent in the current interval
	
	uint curNumAck ;	//number of packets ACKed in the current interval--done on cum. ack state

	uint curAllowable;	//number of packets allowed to be sent this interval 
	Time intervalDeadline;	//when the current interval ends

	boolint firstInterval;	//whether or not it is the first interval for this flow --need to initialize in ackcache init!

	//debug stuff
	uint curNumDrop; //how many pkts dropped in the interval


	uint curPTNumDrop; //how many pkts dropped in the interval (in protocoltransmit)
	uint curCQNumDrop; //how many pkts dropped in the interval (in protocolcheckqueue)

	uint curNumE2ER; //how many e2e acks received
	uint curNumReTx; //how many retransmissions in given interval
	
#endif


#if NEW_E2E_RATE_CONTROL	
	uint prevNumAck ;			//number of packets ACKed at beginning of current interval--done on cum. ack state
	ushort E2ENumAckSkipped;	//the number of times we've consecutively skipped an ACK (for ACK thinning)
	KTIMER E2EAckTimer;			//allows us to set the timeout for ACK thinning 
	KDPC E2EAckTimeoutDpc;		//for firing the timer
	uint E2ENumNewPktAck;		//number of new pkts we received that we had to send an ACK for (cumulative)
	
	uint saveNumAckPrev ;		//number of packets ACKed in the previous interval

	//debug
	uint totalNumSent;			//total pkts sent
	uint totalReTxRcv;			//total pkts received

#endif


#if SLIDING_RATE_LIMIT_ENABLED	
	//rate-limit at source (per flow)
	TimeStruct *SendTimes ;	//linked list of timestamps for send packets
	TimeStruct *lastSendTimes;	//pointer to last item in linked list
	uint numSendTimes;		//size of SendTimes
	TimeStruct *AckTimes;	//linked list of timestamps for acked packets
	TimeStruct *lastAckTimes;	//pointer to last item in linked list
	uint numAckTimes;		//size of AckTimes
	uint numForwarders;		//number of forwarders included in last packet 
	uint virtualLoss;		//virtual loss to next hop
#endif


#if AIMD_RATE_LIMIT_ENABLED	
	//uint prevNumSent ; //number of packets sent in previous interval
	uint curNumSent;	//number of packet sent in the current interval
	//uint prevNumAck;	//number of packets ACKed in the previous interval
	uint curNumAck ;	//number of packets ACKed in the current interval--done on cum. ack state
	Time intervalDeadline;	//when the current interval ends
	uint aimd_prob;			//probability we use to to determine if we should send the packet
	uint prev_aimd_prob ;	//previous aimd_prob
	boolint firstInterval;	//whether or not it is the first interval for this flow --need to initialize in ackcache init!
	uint virtualLoss;		//virtual loss to next hop

	LQSRAckId	prevAckSeqNum;	//cumulative sequence number at beginning of the interval
	AckBitMap	prevAckBitMap;	//bitmap at begining of the interval
#endif

};



struct AckCacheList {
	AckCache *head;		//the head of the linked list.
};

struct MaintBufPacket {
    MaintBufPacket *Next;
    uint RefCnt;
    SRPacket *srp;
    LQSRAckId AckNum;
};

struct MaintBufNode {
    MaintBufNode *Next;

    VirtualAddress Address;
    LQSRIf OutIf;
    LQSRIf InIf;

    //
    // Fields used for transmitting Acknowledgement Requests.
    //

    LQSRAckId NextAckNum;        // Sequence number for next ack request.
    LQSRAckId LastAckNum;        // Last (highest) acknowledged sequence number.
    Time LastAckRcv;            // Time that LackAckNum was received.
    Time FirstAckReq;           // Time that first ack request was sent.
    Time LastAckReq;            // Time that most recent ack request was sent.

    uint NumPackets;
    uint HighWater;

    uint NumAckReqs;            // Number of ack requests sent.
    uint NumFastReqs;           // Number of fast-path ack requests.
    uint NumValidAcks;          // Number of acks received.
    uint NumInvalidAcks;        // Number of invalid acks received.

    MaintBufPacket *MBP;
};

struct MaintBuf {
    KSPIN_LOCK Lock;
    MaintBufNode *MBN;
    uint NumPackets;    // Across all MaintBufNodes.
    uint HighWater;     // Across all MaintBufNodes.
};

//ejr
extern void
updateCachingEntry(AckCache *ac,CachingEntry *CE, VirtualAddress *Forwarder, LQSRAckId seqNum);

extern boolint
checkCachingEntry(AckCache *ac,CachingEntry *CE, SourceRoute *SR, VirtualAddress *Forwarder, LQSRAckId seqNum);

extern AckCache* 
MaintBufFindAckFlow(MiniportAdapter *VA, VirtualAddress *Source, VirtualAddress *Dest, LQSRIf InIf, LQSRIf OutIf,
					VirtualAddress *to) ;

extern void
MaintbufAckTimeout(PKDPC Dpc, void *Context, void *Unused1, void *Unused2);

#if NEW_E2E_RATE_CONTROL
extern void 
MaintbufE2EAckTimeout(PKDPC Dpc, void *Context, void *Unused1, void *Unused2);
#endif

extern void
AckCacheFillAck(AckCache *ac, LQSRAckId *fillSeqNum, AckBitMap *fillBitmap, LQSRAckId rcvdSeqNum); 

extern boolint 
AckCacheDeletePacket(InternalAcknowledgement *Ack, LQSRAckId storeSeqNum) ;

extern boolint 
AckCacheCheckSeqNum(AckCache *AC, LQSRAckId seqNum) ;

extern void 
AckCacheUpdateSeqNum(AckCache *AC, LQSRAckId seqNum);

extern MaintBuf *
MaintBufNew(void);

extern void
MaintBufFree(MiniportAdapter *VA);

extern Time
MaintBufTimer(MiniportAdapter *VA, Time Now);

extern void 
MaintBufRecvAck(MiniportAdapter *VA, const VirtualAddress Address,
                LQSRIf InIf, LQSRIf OutIf, LQSRAckId AckNum);

extern InternalAcknowledgement* 
MaintBufFillInPiggyBackAck(MiniportAdapter *VA, SRPacket *SRP);


extern InternalAcknowledgement* 
MaintBufGetCrossFlowAcks(MiniportAdapter *VA, AckCache *AC, uint Size, uint numAcks);

extern void
MaintBufSendPacket(MiniportAdapter *VA, SRPacket *srp,
                   void (*Complete)(MiniportAdapter *VA, SRPacket *srp,
                                    NDIS_STATUS Status));
extern void
MaintBufRetransmitSendComplete(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status);

extern void
MaintBufStaticSendComplete(MiniportAdapter *VA, NDIS_PACKET *Packet, NDIS_STATUS Status) ;

extern boolint
MaintBufSendAck(MiniportAdapter *VA, SRPacket *srp);

extern void
ACKSendComplete(MiniportAdapter *VA, SRPacket *SRP, NDIS_STATUS Status);

extern void
MaintBufResetStatistics(MiniportAdapter *VA);

//ejr- this wasn't defined here previously, and used to be static
extern MaintBufNode *
MaintBufFindNode(MiniportAdapter *VA,
				 MaintBuf *MB,
                 const VirtualAddress Address,
                 LQSRIf InIf, LQSRIf OutIf);