// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#ifndef __SR_H__
#define __SR_H__

//
// Generic form of internal option representation.
// All internal option formats must match these initial fields.
//
typedef struct _InternalOption InternalOption;
typedef struct _InternalOption {
    InternalOption *Next;
    LQSROption Opt;
};

typedef struct _InternalRouteReply InternalRouteReply;
struct _InternalRouteReply {
    InternalRouteReply *next;
    RouteReply opt;
    // Remember to allocate sizeof(SRAddr) * MAX_SR_LEN extra bytes!
};

typedef struct _InternalRouteRequest InternalRouteRequest;
struct _InternalRouteRequest {
    InternalRouteRequest *next;
    RouteRequest opt;
    // Remember to allocate sizeof(SRAddr) * MAX_SR_LEN extra bytes!
};



typedef struct _InternalSourceRoute InternalSourceRoute;
struct _InternalSourceRoute {
    InternalSourceRoute *next;
    SourceRoute opt;
    // Remember to allocate sizeof(SRAddr) * MAX_SR_LEN extra bytes!
};

typedef struct _InternalRouteError InternalRouteError;
struct _InternalRouteError {
    InternalRouteError *next;
    RouteError opt;
};

typedef struct _InternalAcknowledgementRequest InternalAcknowledgementRequest;
struct _InternalAcknowledgementRequest {
    InternalAcknowledgementRequest *next;
    AcknowledgementRequest opt;
};

typedef struct _InternalAcknowledgement InternalAcknowledgement;
struct _InternalAcknowledgement {
    InternalAcknowledgement *next;
    Acknowledgement opt;
};

typedef struct _InternalInfoRequest InternalInfoRequest;
struct _InternalInfoRequest {
    InternalInfoRequest *next;
    InfoRequest opt;
};

typedef struct _InternalInfoReply InternalInfoReply;
struct _InternalInfoReply {
    InternalInfoReply *next;
    InfoReply opt;
};

typedef struct _InternalProbe InternalProbe;
struct _InternalProbe {
    InternalProbe *Next;
    Probe Opt;
};

typedef struct _InternalProbeReply InternalProbeReply;
struct _InternalProbeReply {
    InternalProbeReply *Next;
    ProbeReply Opt;
};

typedef struct _InternalLinkInfo InternalLinkInfo;
struct _InternalLinkInfo {
    InternalLinkInfo *Next;
    LinkInfo Opt;
};

typedef struct _SRPacket SRPacket;
struct _SRPacket {
    SRPacket *Next;

    PhysicalAddress EtherDest;
    PhysicalAddress EtherSource;
    VirtualAddress Dest;
    VirtualAddress Source;
	uchar IV[LQSR_IV_LENGTH];
    uchar MAC[LQSR_MAC_LENGTH]; //ejr
	SRPacketHolder *RecoveryHolder ; //ejr

    //
    // If you add a new option type, be sure to update
    // SROptionLength and PbackPacketSize as well as
    // SRPacketToPkt, SRPacketFromPkt etc.
    //
    InternalSourceRoute *sr;
	//InternalRecoveryNodes *recovery; //ejr- the local recovery node list (should be very similiar to sr)
    InternalRouteRequest *req;
    InternalRouteReply *rep;
    InternalRouteError *err;
    InternalAcknowledgementRequest *ackreq;
    InternalAcknowledgement *ack;
    InternalInfoRequest *inforeq;
    InternalInfoReply *inforep;
    InternalProbe *Probe;
    InternalProbeReply *ProbeReply;
    InternalLinkInfo *LinkInfo;

    //
    // The encrypted payload carried in this SRPacket.
    // Packet may be NULL, for example for an Ack or a Route Request,
    // in which case PayloadOffset should be zero.
    //
    NDIS_PACKET *Packet;
    uint PacketLength;          // Valid on the receive path.
    uint PayloadOffset;


    ProtocolAdapter *PA;
    void (*FreePacket)(ProtocolAdapter *PA, NDIS_PACKET *Packet);
    void (*FreePacket2)(MiniportAdapter *VA, ProtocolAdapter *PA, NDIS_PACKET *Packet);

    //
    // Used by ProtocolForwardRequest.
    //
    void (*TransmitComplete)(MiniportAdapter *VA, SRPacket *Srp, NDIS_STATUS Status);
    uint ForwardCount;
    NDIS_STATUS ForwardStatus;
	boolint sendCompleteDone ; //set to 0 init., should go to 1 when MaintBufRetransmitSendComplete is done.
	boolint sendCompleteShouldDelete; //set to 0 init.... should go to 1 when MaintBufRetransmitSendComplete
										//is responsible for deleting the SRPacket.
	boolint ndisMSendCompleteDone ;		//set to 1 after we call NdisMSendComplete on the packet.


    NDIS_PACKET *OriginalPacket;		//this will be packet we get from higher layer.
	void *dataPtr;						//ptr to data we allocate for our cloned packet. 
	MiniportAdapter *VA;
	NDIS_PACKET *copiedPacket;			//have to free this as well.

	char *MBNnexthop ; //need this for retransmission timer in protocoltransmit, of type: MaintBufNode
    InternalSourceRoute *old_sr; //when we forward data, keep old source route to determine progress

	Time updateTime;//to keep track when the packet was last updated.


};


extern SRPacket*
FillSRPacket(SRPacket *SRP, LQSRHeader *LQSR);

extern void 
SRPacketFree(SRPacket *Srp);

extern void
SRFreeOptionList(InternalOption *List);

extern void
InsertOptions(LQSROption **Walk, void *Field, uchar Type);

extern void
InsertFakeOptions(LQSROption **Walk, void *Field, uchar Type);

extern uint
SROptionListLength(InternalOption *List);

extern NDIS_STATUS
SRPacketFromPkt(MiniportAdapter *VA, NDIS_PACKET *Packet, SRPacket **pSRP);

extern NDIS_STATUS
SRPacketToPkt(MiniportAdapter *VA, const SRPacket *Srp, NDIS_PACKET **pPacket);

extern boolint 
checkSRPbitmap(ushort bitmap, uint index) ;

extern void 
setSRPbitmap(ushort *bitmap, uint index) ;


extern NDIS_STATUS 
SRRetransmissionPacketToPkt(MiniportAdapter *VA, const SRPacket *SRP, NDIS_PACKET **pPacket);

extern uint
SROptionLength(const SRPacket *SRP);

extern NDIS_STATUS
CheckPacket(MiniportAdapter *VA, NDIS_PACKET *Packet);

#endif /* __SR_H__ */
